from flask import current_app

from ewoksjob.client import submit
from ewoksjob.client.local import submit as submit_local
from . import resource
from .. import api


class Workflow(resource.JsonResource):
    RESOURCE_TYPE = "workflow"

    def get_identifier(
        self, resource: resource.ResourceContentType
    ) -> resource.ResourceIdentifierType:
        return resource["graph"]["id"]

    @api.get_resource("workflow")
    def get(self, identifier: resource.ResourceIdentifierType) -> resource.ResponseType:
        return self.load_resource(identifier)

    @api.put_resource("workflow")
    def put(
        self, identifier: resource.ResourceIdentifierType, **resource
    ) -> resource.ResponseType:
        return self.save_resource(
            resource, error_on_missing=True, identifier=identifier
        )

    @api.delete_resource("workflow")
    def delete(
        self, identifier: resource.ResourceIdentifierType
    ) -> resource.ResponseType:
        return self.delete_resource(identifier)


class Workflows(resource.JsonResource):
    RESOURCE_TYPE = "workflow"

    def get_identifier(
        self, resource: resource.ResourceContentType
    ) -> resource.ResourceIdentifierType:
        return resource["graph"]["id"]

    @api.list_resource_identifiers("workflow")
    def get(self) -> resource.ResponseType:
        return self.list_resource_identifiers()

    @api.post_resource("workflow")
    def post(self, **resource) -> resource.ResponseType:
        return self.save_resource(resource, error_on_exists=True)


class Description(resource.JsonResource):
    RESOURCE_TYPE = "workflow"

    def get_identifier(
        self, resource: resource.ResourceContentType
    ) -> resource.ResourceIdentifierType:
        return resource["graph"]["id"]

    @api.list_resource_descriptions("workflow")
    def get(self) -> resource.ResponseType:
        return self.list_resource_descriptions()


class Execute(resource.JsonResource):
    RESOURCE_TYPE = "workflow"

    def get_identifier(
        self, resource: resource.ResourceContentType
    ) -> resource.ResourceIdentifierType:
        return resource["graph"]["id"]

    @api.execute_resource("workflow")
    def post(
        self,
        identifier: resource.ResourceIdentifierType,
        execute_arguments=None,
        worker_options=None,
    ):
        graph, error_code = self.load_resource(identifier)
        if error_code != 200:
            return graph, error_code

        if execute_arguments is None:
            execute_arguments = dict()
        if worker_options is None:
            submit_kwargs = dict()
        else:
            submit_kwargs = dict(worker_options)
        submit_kwargs["args"] = (graph,)
        submit_kwargs["kwargs"] = execute_arguments

        ewoks_config = current_app.config.get("EWOKS")
        if ewoks_config:
            execinfo = execute_arguments.setdefault("execinfo", dict())
            handlers = execinfo.setdefault("handlers", list())
            for handler in ewoks_config.get("handlers", list()):
                if handler not in handlers:
                    handlers.append(handler)

        if current_app.config.get("CELERY") is None:
            future = submit_local(**submit_kwargs)
        else:
            future = submit(**submit_kwargs)
        return self.make_response(200, job_id=future.task_id)



"""Flask resources that implement the REST API using json as backend.
"""

from flask_restful import Api
from flask_apispec import FlaskApiSpec
from . import tasks
from . import workflows
from ..utils import register_resource


def add_resources(api: Api, apidoc: FlaskApiSpec):
    # Save/load/execute workflows
    register_resource(workflows.Workflows, "/workflows", api, apidoc)
    register_resource(workflows.Description, "/workflows/descriptions", api, apidoc)
    register_resource(workflows.Workflow, "/workflow/<string:identifier>", api, apidoc)
    register_resource(workflows.Execute, "/execute/<string:identifier>", api, apidoc)

    # Save/load tasks
    register_resource(tasks.Tasks, "/tasks", api, apidoc)
    register_resource(tasks.Task, "/task/<string:identifier>", api, apidoc)
    register_resource(tasks.Descriptions, "/tasks/descriptions", api, apidoc)
    register_resource(tasks.DiscoverTasks, "/tasks/discover", api, apidoc)



from typing import Type
from flask_restful import Api
from flask_apispec import FlaskApiSpec
from flask_restful import Resource as _Resource
from flask_apispec import MethodResource


class Resource(MethodResource, _Resource):
    pass


def register_resource(
    resource: Type[Resource], path: str, api: Api, apidoc: FlaskApiSpec
):
    api.add_resource(resource, path)
    apidoc.register(resource)



from typing import Tuple
from flask import current_app
from ..utils import Resource
from . import utils
from .utils import ResourceUrlType
from .utils import ResourceContentType
from .utils import ResourceIdentifierType

ResponseType = Tuple[dict, int]


class BinaryResource(Resource):
    """Base class without end points"""

    RESOURCE_TYPE = NotImplemented

    @property
    def root_url(self) -> ResourceUrlType:
        return utils.root_url(
            current_app.config.get("RESOURCE_DIRECTORY"), self.RESOURCE_TYPE + "s"
        )

    def save_resource(
        self,
        identifier: ResourceIdentifierType,
        resource: ResourceContentType,
        error_on_exists: bool = False,
        error_on_missing: bool = False,
    ) -> ResponseType:
        """
        200: OK
        400: bad request (fails due to a client error)
        403: forbidden
        404: not found (`error_on_missing=True`)
        409: already exists  (`error_on_exists=True`)
        """
        root_url = self.root_url
        exists = utils.resource_exists(root_url, identifier)
        if error_on_exists and exists:
            return self.make_response(
                409,
                message=f"{self.RESOURCE_TYPE.capitalize()} '{identifier}' already exists.",
                identifier=identifier,
            )

        if error_on_missing and not exists:
            return self.make_response(
                404,
                message=f"{self.RESOURCE_TYPE.capitalize()} '{identifier}' is not found.",
                identifier=identifier,
            )

        try:
            utils.save_resource(root_url, identifier, resource)
        except PermissionError:
            return self.make_response(
                403,
                message=f"No permission to write {self.RESOURCE_TYPE} '{identifier}.'",
                identifier=identifier,
            )

        return resource, 200

    def load_resource(
        self,
        identifier: ResourceIdentifierType,
    ) -> ResponseType:
        """
        200: OK
        403: forbidden
        404: not found
        """
        try:
            return utils.load_resource(self.root_url, identifier), 200
        except PermissionError:
            return self.make_response(
                403,
                message=f"No permission to read {self.RESOURCE_TYPE} '{identifier}'",
                identifier=identifier,
            )
        except FileNotFoundError:
            return self.make_response(
                404,
                message=f"{self.RESOURCE_TYPE.capitalize()} '{identifier}' is not found.",
                identifier=identifier,
            )

    def delete_resource(self, identifier: ResourceIdentifierType) -> ResponseType:
        """
        200: OK
        403: forbidden
        """
        try:
            utils.delete_resource(self.root_url, identifier)
        except PermissionError:
            return self.make_response(
                403,
                message=f"No permission to delete {self.RESOURCE_TYPE} '{identifier}'.",
                identifier=identifier,
            )
        return self.make_response(200, identifier=identifier)

    def list_resource_identifiers(self) -> ResponseType:
        """
        200: OK
        """
        body = {"identifiers": list(utils.resource_identifiers(self.root_url))}
        return body, 200

    def list_resource_content(self) -> ResponseType:
        """
        200: OK
        """
        body = {"items": list(utils.resources(self.root_url))}
        return body, 200

    def get_identifier(self, resource: ResourceContentType) -> ResourceIdentifierType:
        raise NotImplementedError

    @classmethod
    def make_response(cls, code: int, **body) -> ResponseType:
        body["type"] = cls.RESOURCE_TYPE
        return body, code



import time
from ewokscore.tests.examples.graphs import get_graph


def test_execute_with_celery(celery_exec_client):
    _test_execute(*celery_exec_client)


def test_execute_without_celery(local_exec_client):
    _test_execute(*local_exec_client)


def test_new_client_new_events(local_exec_client):
    client, sclient = local_exec_client
    _test_execute(client, sclient)
    sclient.disconnect()
    sclient.connect()
    time.sleep(1)
    assert not sclient.get_received()


def _test_execute(client, sclient):
    graph_name, expected = upload_graph(client)
    response = client.post(f"/execute/{graph_name}")
    assert response.status_code == 200, response.get_json()

    n = 2 * (len(expected) + 2)
    events = get_events(sclient, n)
    _assert_events(response, events, expected)
    return n


def upload_graph(client):
    graph_name = "acyclic1"
    graph, expected = get_graph(graph_name)
    response = client.post("/workflows", json=graph)
    assert response.status_code == 200, response.get_json()
    return graph_name, expected


def get_events(sclient, nevents, timeout=3):
    t0 = time.time()
    events = list()
    while True:
        new_events = sclient.get_received()
        events.extend(new_events)
        if len(events) == nevents:
            break
        time.sleep(0.1)
        if time.time() - t0 > timeout:
            break

    ewoks_events = list()
    for flask_event in events:
        ewoks_events.extend(flask_event["args"])
    return ewoks_events


def _assert_events(response, events, expected):
    n = 2 * (len(expected) + 2)
    assert len(events) == n

    job_id = response.get_json()["job_id"]
    for event in events:
        assert event["job_id"] == job_id
        if event["node_id"]:
            assert event["node_id"] in expected



from contextlib import contextmanager
from typing import Iterator, Optional

from flask import current_app

from ewoksjob.events.readers import EwoksEventReader
from ewoksjob.events.readers import instantiate_reader


def reader() -> Optional[EwoksEventReader]:
    cfg = current_app.config.get("EWOKS", dict())
    handlers = cfg.get("handlers", list())
    argmap = {"uri": "url"}
    for name in ("Redis", "Sqlite3", None):
        for handler in handlers:
            if name is None or name in handler["class"]:
                arguments = handler.get("arguments", list())
                arguments = {
                    argmap.get(arg["name"], arg["name"]): arg["value"]
                    for arg in arguments
                }
                return instantiate_reader(**arguments)

    current_app.logger.warning("Configure ewoks event handlers")
    return None


@contextmanager
def reader_context() -> Iterator[Optional[EwoksEventReader]]:
    r = reader()
    try:
        yield r
    finally:
        if r is not None:
            r.close()



def test_single_icon_png(rest_client, png_icons):
    assert_single_icon(rest_client, png_icons, ".png")


def test_single_icon_svg(rest_client, svg_icons):
    assert_single_icon(rest_client, svg_icons, ".svg")


def test_multiple_icons_png(rest_client, png_icons, default_icon_identifiers):
    assert_multiple_icons(rest_client, png_icons, ".png", default_icon_identifiers)


def test_multiple_icons_svg(rest_client, svg_icons, default_icon_identifiers):
    assert_multiple_icons(rest_client, svg_icons, ".svg", default_icon_identifiers)


def assert_single_icon(rest_client, icons, ext):
    identifier = "icon" + ext

    response = rest_client.get(f"/icon/{identifier}")
    assert response.status_code == 404

    icon1a = icons[0]
    response = rest_client.post(f"/icon/{identifier}", json=icon1a)
    data = response.get_json()
    assert response.status_code == 200, data
    assert data == icon1a

    response = rest_client.get(f"/icon/{identifier}")
    data = response.get_json()
    assert response.status_code == 200, data
    assert data == icon1a

    icon1b = icons[1]
    response = rest_client.put(f"/icon/{identifier}", json=icon1b)
    data = response.get_json()
    assert response.status_code == 200, data
    assert data == icon1b

    response = rest_client.get(f"/icon/{identifier}")
    data = response.get_json()
    assert response.status_code == 200, data
    assert data == icon1b

    response = rest_client.delete(f"/icon/{identifier}")
    data = response.get_json()
    assert response.status_code == 200
    assert data == {"identifier": identifier}

    response = rest_client.delete(f"/icon/{identifier}")
    data = response.get_json()
    assert response.status_code == 200
    assert data == {"identifier": identifier}

    response = rest_client.get(f"/icon/{identifier}")
    data = response.get_json()
    assert response.status_code == 404
    expected = {
        "identifier": identifier,
        "message": f"Icon '{identifier}' is not found.",
        "type": "icon",
    }
    assert data == expected


def assert_multiple_icons(rest_client, icons, ext, existing):
    expected = list(existing)
    for i, icon in enumerate(icons):
        identifier = f"icon{i}{ext}"
        expected.append(identifier)
        response = rest_client.post(f"/icon/{identifier}", json=icon)
        data = response.get_json()
        assert response.status_code == 200, data
        assert data == icon

    response = rest_client.get("/icons")
    data = response.get_json()
    assert response.status_code == 200, data
    assert set(data["identifiers"]) == set(expected)



import pytest
import shutil
from pathlib import Path
from ..resources.binary import utils as binary_utils
from ..resources.json import utils as json_utils
from ..resources.data import DEFAULT_ROOT


def test_icon_fallback(tmpdir, default_icon_identifiers):
    category = "icons"
    existing = "default.png"
    nonexisting = "nonexisting.png"
    user = "user.png"
    assert_resource_fallback(
        binary_utils,
        tmpdir,
        category,
        existing,
        nonexisting,
        user,
        ".png",
        default_icon_identifiers,
    )


def test_workflows_fallback(tmpdir, default_workflow_identifiers):
    category = "workflows"
    existing = "demo"
    nonexisting = "nonexisting"
    user = "user"
    assert_resource_fallback(
        json_utils,
        tmpdir,
        category,
        existing,
        nonexisting,
        user,
        ".json",
        default_workflow_identifiers,
    )


def assert_resource_fallback(
    utils, user_root, category, existing, nonexisting, user, extension, all_resources
):
    user_root = Path(user_root).resolve()
    user_icons_root = utils.root_url(user_root, category)
    default_icons_root = utils.root_url(DEFAULT_ROOT, category)

    source = (default_icons_root / existing).with_suffix(extension)
    dest = (user_icons_root / user).with_suffix(extension)
    user_icons_root.mkdir()
    shutil.copyfile(source, dest)

    # File only in default root
    utils.load_resource(user_icons_root, existing)

    # File neither in  default nore in user root
    with pytest.raises(FileNotFoundError):
        utils.load_resource(user_icons_root, nonexisting)

    # File only in user root
    utils.load_resource(user_icons_root, user)

    # Find all resources
    found = sorted(utils.resource_identifiers(user_icons_root))
    assert found == sorted(all_resources + [user])

    # Delete user resource
    utils.delete_resource(user_icons_root, user)

    found = sorted(utils.resource_identifiers(user_icons_root))
    assert found == sorted(all_resources)



from . import resource
from .. import api


class Icon(resource.BinaryResource):
    RESOURCE_TYPE = "icon"

    @api.get_resource("icon")
    def get(
        self, identifier: resource.ResourceIdentifierType
    ) -> resource.ResourceIdentifierType:
        return self.load_resource(identifier)

    @api.delete_resource("icon")
    def delete(
        self, identifier: resource.ResourceIdentifierType
    ) -> resource.ResponseType:
        return self.delete_resource(identifier)

    @api.put_resource("icon")
    def put(
        self, identifier: resource.ResourceIdentifierType, **resource
    ) -> resource.ResponseType:
        return self.save_resource(identifier, resource, error_on_missing=True)

    @api.post_resource("icon")
    def post(
        self, identifier: resource.ResourceIdentifierType, **resource
    ) -> resource.ResponseType:
        return self.save_resource(identifier, resource, error_on_exists=True)


class Icons(resource.BinaryResource):
    RESOURCE_TYPE = "icon"

    @api.list_resource_identifiers("icon")
    def get(self) -> resource.ResponseType:
        return self.list_resource_identifiers()



from typing import List

from ewoksutils import event_utils

from ..utils import Resource
from .. import api
from ...events.ewoks_events import reader_context


class ExecutionEvents(Resource):
    @api.get_ewoks_events()
    def get(self, **filters) -> List[List[dict]]:
        jobs = list()
        job = None
        job_id = None
        with reader_context() as reader:
            if reader is None:
                raise RuntimeError("server not configured for ewoks events")
            for event in reader.get_events(**filters):
                if event["job_id"] != job_id:
                    job_id = event["job_id"]
                    job = list()
                    jobs.append(job)
                if "engine" in event_utils.FIELD_TYPES:
                    event["binding"] = event["engine"]
                job.append(event)
        return {"jobs": jobs}, 200



import os
import sys
import json
from pprint import pformat
from typing import Optional, Tuple

import argparse
import logging
from contextlib import contextmanager

import flask
from flask_cors import CORS
from flask_socketio import SocketIO
from flask_restful import Api
from flask_apispec import FlaskApiSpec

from celery import current_app as current_celery_app
from ewoksjob.client.local import pool_context

try:
    from ewoksweb.serverutils import get_static_root
except ImportError:

    def get_static_root():
        return "static"


from .resources import add_resources
from .events.websocket import add_events


def create_app(**config) -> Tuple[flask.Flask, Api, FlaskApiSpec]:
    app = flask.Flask(__name__, static_folder=get_static_root(), static_url_path="")
    CORS(app)
    configure_app(app, **config)
    apidoc = add_apidoc(app)
    api = add_api(app, apidoc)
    return app, api, apidoc


def add_api(app: flask.Flask, apidoc: FlaskApiSpec) -> Api:
    api = Api(app)
    add_resources(api, apidoc)
    return api


def configure_app(app: flask.Flask, configuration: Optional[str] = None, **config):
    app.config["CORS_HEADERS"] = "Content-Type"
    filename = os.environ.get("EWOKSSERVER_SETTINGS")
    if configuration:
        filename = configuration
    if filename:
        filename = os.path.relpath(filename, app.config.root_path)
        app.config.from_pyfile(filename, silent=False)
    if config:
        config = {k.upper(): v for k, v in config.items() if v is not None}
        app.config.update(config)
    if app.config.get("CELERY"):
        current_celery_app.conf.update(app.config["CELERY"])


def print_config(app: flask.Flask):
    resourcedir = app.config.get("RESOURCE_DIRECTORY")
    if not resourcedir:
        resourcedir = "."
    print(f"\nRESOURCE DIRECTORY:\n {os.path.abspath(resourcedir)}\n")

    adict = app.config.get("CELERY")
    if adict:
        print(f"\nCELERY:\n {pformat(adict)}\n")
    else:
        print("\nCELERY:\n Not configured (local workflow execution)\n")

    adict = app.config.get("EWOKS")
    if adict:
        print(f"\nEWOKS:\n {pformat(adict)}\n")
    else:
        print("\nEWOKS:\n Not configured (no ewoks execution events)\n")


def print_serve_message(app, port: Optional[int] = None) -> None:
    host = "127.0.0.1"
    if port is None:
        server_name = app.config["SERVER_NAME"]
        if server_name and ":" in server_name:
            port = int(server_name.rsplit(":", 1)[1])
        else:
            port = 5000
    print("\nTo start editing workflows, open this link in a browser:\n")
    print(f"    http://{host}:{port}\n")


def set_log_level(app: Optional[flask.Flask] = None, log_level=logging.WARNING):
    logging.basicConfig(level=log_level)
    if app is not None:
        app.logger.setLevel(log_level)


def add_socket(app: flask.Flask) -> SocketIO:
    socketio = SocketIO(app, cors_allowed_origins="*")
    add_events(socketio)
    return socketio


def add_apidoc(app: flask.Flask) -> FlaskApiSpec:
    app.config.update(
        {
            "APISPEC_TITLE": "ewoks",
            # "APISPEC_OAS_VERSION": "3.0.3",
            "APISPEC_SWAGGER_URL": "/swagger/",
            "APISPEC_SWAGGER_UI_URL": "/swagger-ui/",
        }
    )
    return FlaskApiSpec(app)


def save_apidoc(apidoc: FlaskApiSpec, filename: str) -> None:
    save_spec_dir = os.path.dirname(filename)
    if save_spec_dir:
        os.makedirs(save_spec_dir, exist_ok=True)
    with open(filename, "w") as f:
        json.dump(apidoc.spec.to_dict(), f)


def run_app(
    app: flask.Flask, socketio: Optional[SocketIO] = None, port: int = 5000
) -> None:
    with run_context(app):
        if socketio is None:
            app.run(port=port)
        else:
            # allow_unsafe_werkzeug=True, see MR1877
            socketio.run(app, port=port)


@contextmanager
def run_context(app: flask.Flask):
    with app.app_context():
        if app.config.get("CELERY") is None:
            with pool_context():
                yield
        else:
            yield


def main(argv=None):
    if argv is None:
        argv = sys.argv

    parser = argparse.ArgumentParser(description="REST API for Ewoks")
    parser.add_argument(
        "-l",
        "--log",
        dest="log_level",
        type=str.upper,
        default="WARNING",
        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
        help="Set the logging level",
    )
    parser.add_argument(
        "-p",
        "--port",
        dest="port",
        type=int,
        default=5000,
        help="Port number",
    )
    parser.add_argument(
        "-c",
        "--config",
        dest="configuration",
        type=str,
        default=None,
        help="Path to a python script (equivalent to the environment variable 'EWOKSSERVER_SETTINGS')",
    )
    parser.add_argument(
        "-d",
        "--dir",
        dest="resource_directory",
        type=str,
        default=None,
        help="Root directory for resources (e.g. workflows, tasks, icons descriptions)",
    )
    parser.add_argument(
        "-s",
        "--spec-filename",
        dest="spec_filename",
        type=str,
        help="Save the Swagger docs as JSON",
    )
    parser.add_argument(
        "--without-events",
        action="store_true",
        help="Without websocket events",
    )

    args = parser.parse_args(argv[1:])
    log_level = getattr(logging, args.log_level)

    app, _, apidoc = create_app(
        configuration=args.configuration, resource_directory=args.resource_directory
    )
    if args.spec_filename:
        save_apidoc(apidoc, args.spec_filename)
        return
    if args.without_events:
        socketio = None
    else:
        socketio = add_socket(app)
    set_log_level(log_level=log_level)

    print_config(app)
    print_serve_message(app, port=args.port)
    run_app(app, socketio=socketio, port=args.port)


if __name__ == "__main__":
    sys.exit(main())



"""Flask resources that implement the REST API using the ewoks event database as backend.
"""

from flask_restful import Api
from flask_apispec import FlaskApiSpec
from . import resource
from ..utils import register_resource


def add_resources(api: Api, apidoc: FlaskApiSpec):
    register_resource(resource.ExecutionEvents, "/execution/events", api, apidoc)



import os
from glob import glob
from typing import List

_SCRIPT_DIR = os.path.abspath(os.path.dirname(__file__))


def resource_filenames() -> List[str]:
    files = glob(os.path.join(_SCRIPT_DIR, "*.*"))
    files = [filename for filename in files if filename != __file__]
    return files



import time
from ..events.websocket import is_running


def test_websocket_connection_local(local_exec_client):
    _, sclient = local_exec_client
    _test_websocket_connection(sclient)


def test_websocket_connection_celery(celery_exec_client):
    _, sclient = celery_exec_client
    _test_websocket_connection(sclient)


def _test_websocket_connection(sclient):
    assert sclient.is_connected()
    _assert_eventloop_is_running(True)
    sclient.disconnect()
    _assert_eventloop_is_running(False)
    sclient.connect()
    _assert_eventloop_is_running(True)


def _assert_eventloop_is_running(running, timeout=3):
    t0 = time.time()
    while True:
        if is_running() == running:
            return
        time.sleep(0.1)
        if time.time() - t0 > timeout:
            raise TimeoutError



"""Flask resources that implement the REST API using binary as backend.
"""

from flask_restful import Api
from flask_apispec import FlaskApiSpec
from . import icons
from ..utils import register_resource


def add_resources(api: Api, apidoc: FlaskApiSpec):
    # Save/load icons
    register_resource(icons.Icons, "/icons", api, apidoc)
    register_resource(icons.Icon, "/icon/<string:identifier>", api, apidoc)



import os
from datetime import datetime
from .test_execute import upload_graph
from .test_execute import get_events


def test_get_execution_events(local_exec_client):
    client, sclient = local_exec_client

    graph_name, expected = upload_graph(client)
    nevents = 0
    nevents_per_exec = 2 * (len(expected) + 2)

    # Test no events (nothing has been executed)
    response = client.get("/execution/events")
    assert response.status_code == 200
    data = response.get_json()
    assert data == {"jobs": list()}

    # Execute workflow
    response = client.post(f"/execute/{graph_name}")
    data = response.get_json()
    assert response.status_code == 200, data
    job_id1 = data["job_id"]
    nevents += nevents_per_exec

    # Wait until all events have been received over the websocket
    events1 = get_events(sclient, nevents, timeout=3)

    # Query should return the same a what was recieved over the websocket
    response = client.get("/execution/events")
    assert response.status_code == 200
    events = response.get_json()["jobs"]
    assert len(events) == 1
    assert events[0] == events1

    response = client.get(f"/execution/events?job_id={job_id1}")
    assert response.status_code == 200
    events = response.get_json()["jobs"]
    assert len(events) == 1
    assert events[0] == events1

    response = client.get("/execution/events?context=job")
    assert response.status_code == 200
    events = response.get_json()["jobs"]
    assert len(events) == 1
    assert len(events[0]) == 2

    dtmid = datetime.now().astimezone()

    # Execute workflow
    response = client.post(f"/execute/{graph_name}")
    data = response.get_json()
    assert response.status_code == 200, data
    job_id2 = data["job_id"]
    nevents += nevents_per_exec

    # Wait until all events have been received over the websocket
    events2 = get_events(sclient, nevents_per_exec, timeout=3)

    response = client.get("/execution/events")
    assert response.status_code == 200
    events = response.get_json()["jobs"]
    assert len(events) == 2
    assert events[0] == events1
    assert events[1] == events2

    response = client.get(f"/execution/events?job_id={job_id2}")
    assert response.status_code == 200
    events = response.get_json()["jobs"]
    assert len(events) == 1
    assert events[0] == events2

    response = client.get("/execution/events?context=job")
    assert response.status_code == 200
    events = response.get_json()["jobs"]
    assert len(events) == 2
    assert len(events[0]) == 2
    assert len(events[1]) == 2

    if os.name == "nt":
        return  # TODO: time filtering fails on windows

    # Test time Query
    midtime = dtmid.isoformat().replace("+", "%2b")
    response = client.get(f"/execution/events?endtime={midtime}")
    assert response.status_code == 200
    events = response.get_json()["jobs"]
    assert len(events) == 1
    assert events[0] == events1

    response = client.get(f"/execution/events?starttime={midtime}")
    assert response.status_code == 200
    events = response.get_json()["jobs"]
    assert len(events) == 1
    assert events[0] == events2



from pathlib import Path
from typing import List
import pytest
from ewoksserver.server import create_app
from ewoksserver.server import run_context
from ewoksserver.server import add_socket
from ewoksjob.client.local import pool_context
from ewoksjob.tests.conftest import celery_config  # noqa F401
from ewoksjob.tests.conftest import celery_includes  # noqa F401

from .data import resource_filenames
from ..resources.binary.utils import _load_url
from ..resources.data import DEFAULT_ROOT


@pytest.fixture
def rest_client(tmpdir):
    """Client to the REST server (no execution)."""
    app, *_ = create_app(resource_directory=str(tmpdir))
    with run_context(app):
        with app.test_client() as client:
            yield client


@pytest.fixture()
def ewoks_handlers(tmpdir):
    uri = f"file:{tmpdir / 'ewoks_events.db'}"
    return [
        {
            "class": "ewokscore.events.handlers.Sqlite3EwoksEventHandler",
            "arguments": [{"name": "uri", "value": uri}],
        }
    ]


@pytest.fixture
def local_exec_client(tmpdir, ewoks_handlers):
    """Client to the REST server and websocket (execution with process pool)."""
    ewoks_config = {"handlers": ewoks_handlers}
    app, *_ = create_app(resource_directory=str(tmpdir), ewoks=ewoks_config)
    socketio = add_socket(app)
    with run_context(app):
        with pool_context():
            with app.test_client() as client:
                sclient = socketio.test_client(app, flask_test_client=client)
                yield client, sclient
                sclient.disconnect()


@pytest.fixture
def celery_exec_client(tmpdir, celery_session_worker, ewoks_handlers):
    """Client to the REST server and websocket (execution with celery)."""
    ewoks_config = {"handlers": ewoks_handlers}
    app, *_ = create_app(
        resource_directory=str(tmpdir), celery=dict(), ewoks=ewoks_config
    )
    socketio = add_socket(app)
    with run_context(app):
        with app.test_client() as client:
            sclient = socketio.test_client(app, flask_test_client=client)
            yield client, sclient
            sclient.disconnect()


@pytest.fixture
def png_icons():
    filenames = resource_filenames()
    return [_load_url(filename) for filename in filenames if filename.endswith(".png")]


@pytest.fixture
def svg_icons():
    filenames = resource_filenames()
    return [_load_url(filename) for filename in filenames if filename.endswith(".svg")]


@pytest.fixture(scope="session")
def default_icon_identifiers() -> List[Path]:
    return [
        url.name
        for url in (DEFAULT_ROOT / "icons").iterdir()
        if not url.name.startswith("__")
    ]


@pytest.fixture(scope="session")
def default_workflow_identifiers() -> List[Path]:
    return [
        url.stem
        for url in (DEFAULT_ROOT / "workflows").iterdir()
        if url.suffix == ".json"
    ]


@pytest.fixture(scope="session")
def default_task_identifiers() -> List[Path]:
    return [
        url.stem for url in (DEFAULT_ROOT / "tasks").iterdir() if url.suffix == ".json"
    ]



import base64
import mimetypes
from urllib import request
import logging
from pathlib import Path
from typing import Iterator, Union
from ..data import DEFAULT_ROOT


ResourceIdentifierType = str
ResourceUrlType = Path
ResourceContentType = dict

_logger = logging.getLogger(__name__)


def root_url(root_url: Union[str, Path, None], category: str) -> ResourceUrlType:
    if not root_url:
        root_url = Path(".")
    elif isinstance(root_url, str):
        root_url = Path(root_url)
    return root_url / category


def resource_identifiers(root: ResourceUrlType) -> Iterator[ResourceIdentifierType]:
    for url in _resource_urls(root):
        yield _url_to_identifier(url)


def resources(root: ResourceUrlType) -> Iterator[ResourceContentType]:
    for url in _resource_urls(root):
        yield _load_url(url)


def resource_exists(root: ResourceUrlType, identifier: ResourceIdentifierType) -> bool:
    for root in [root, _default_root_url(root)]:
        if _identifier_to_url(root, identifier).exists():
            return True
    return False


def _default_root_url(root: ResourceUrlType) -> ResourceUrlType:
    return DEFAULT_ROOT / root.name


def _resource_urls(root: ResourceUrlType) -> Iterator[ResourceUrlType]:
    for root in [root, _default_root_url(root)]:
        if not root.exists():
            continue
        for url in root.iterdir():
            if _is_resource(url):
                yield url


def _is_resource(url: ResourceUrlType) -> bool:
    return (
        url.is_file() and not url.name.startswith(".") and not url.name.endswith(".py")
    )


def save_resource(
    root: ResourceUrlType,
    identifier: ResourceIdentifierType,
    resource: ResourceContentType,
):
    url = _identifier_to_url(root, identifier)
    _save_url(url, resource)


def load_resource(
    root: ResourceUrlType, identifier: ResourceIdentifierType
) -> ResourceContentType:
    url = _identifier_to_url(root, identifier)
    try:
        return _load_url(url)
    except FileNotFoundError:
        url = _identifier_to_url(DEFAULT_ROOT / root.name, identifier)
        return _load_url(url)


def delete_resource(root: ResourceUrlType, identifier: ResourceIdentifierType) -> None:
    url = _identifier_to_url(root, identifier)
    _delete_url(url)


def _delete_url(url: ResourceUrlType) -> ResourceContentType:
    if url.exists():
        _logger.debug("Delete file '%s'", url)
        url.unlink()


def _identifier_to_url(root: ResourceUrlType, identifier: ResourceIdentifierType):
    return root / identifier


def _url_to_identifier(url: ResourceUrlType) -> ResourceIdentifierType:
    return url.name


def _load_url(url: ResourceUrlType) -> ResourceContentType:
    mimetype, encoding = mimetypes.guess_type(url, strict=True)
    if mimetype is None:
        raise ValueError(f"Cannot derive mime type from '{url}'")

    try:
        with open(url, "rb") as f:
            data = f.read()
    except FileNotFoundError:
        _logger.error(f"'{url}' not found")
        raise

    if not encoding:
        encoding = "base64"
        data = base64.b64encode(data).decode()

    return {"data_url": f"data:{mimetype};{encoding},{data}"}


def _save_url(url: ResourceUrlType, resource: ResourceContentType) -> None:
    _logger.debug("Save file '%s'", url)
    url.parent.mkdir(parents=True, exist_ok=True)
    with request.urlopen(resource["data_url"]) as f:
        data = f.read()
    with open(url, "wb") as f:
        f.write(data)



"""Flask resources that implement the REST API.
"""

from flask import current_app
from flask_restful import Api
from flask_apispec import FlaskApiSpec, doc, marshal_with

from . import json
from . import binary
from . import events
from . import utils


class Home(utils.Resource):
    @doc(summary="Ewoks workflow React client")
    @marshal_with(None, code=200)
    def get(self):
        return current_app.send_static_file("index.html")


def add_resources(api: Api, apidoc: FlaskApiSpec):
    """Currently only one resource backend is supported: file system"""
    utils.register_resource(Home, "/", api, apidoc)
    json.add_resources(api, apidoc)
    binary.add_resources(api, apidoc)
    events.add_resources(api, apidoc)



from datetime import datetime
import threading
from typing import Optional

import flask
from flask import copy_current_request_context
from flask_socketio import SocketIO
from flask_socketio import emit

from .ewoks_events import reader_context

fversion = tuple(map(int, flask.__version__.split(".")))[:2]
if fversion < (2, 2):
    from flask.globals import _app_ctx_stack
else:
    _app_ctx_stack = None


def copy_current_app_context(fn):
    if _app_ctx_stack is None:
        return fn

    app_context = _app_ctx_stack.top

    def wrapper(*args, **kwargs):
        with app_context:
            return fn(*args, **kwargs)

    return wrapper


def add_events(socketio: SocketIO):
    socketio.on("connect")(connect)
    socketio.on("disconnect")(disconnect)


def connect():
    _EMITTER.connect()


def disconnect():
    _EMITTER.disconnect()


def is_running():
    return _EMITTER.is_running()


class EwoksEventEmitter:
    def __init__(self) -> None:
        self._stop_event = threading.Event()
        self._thread = None
        self._counter = 0

    def connect(self) -> None:
        self._counter += 1
        self.start()

    def disconnect(self) -> None:
        self._counter = max(self._counter - 1, 0)
        if self._counter == 0:
            self.stop(timeout=3)

    def is_running(self) -> bool:
        return self._is_running(self._thread)

    @staticmethod
    def _is_running(thread: Optional[threading.Thread] = None) -> bool:
        return thread is not None and thread.is_alive()

    def start(self) -> None:
        if self.is_running():
            return

        # Flask context's have thread affinity
        @copy_current_request_context
        @copy_current_app_context
        def main():
            self._main()

        self._stop_event.clear()
        thread = threading.Thread(target=main, daemon=True)
        thread.start()
        self._thread = thread

    def stop(self, timeout: Optional[float] = None) -> None:
        thread = self._thread
        if not self._is_running(thread):
            return
        self._stop_event.set()
        thread.join(timeout=timeout)

    def _main(self) -> None:
        try:
            with reader_context() as reader:
                if reader is None:
                    return
                starttime = datetime.now().astimezone()
                for event in reader.wait_events(
                    starttime=starttime, stop_event=self._stop_event
                ):
                    if self._stop_event.is_set():
                        break
                    emit("Executing", event, broadcast=True)
        finally:
            self._thread = None


_EMITTER = EwoksEventEmitter()



"""Defines the REST API.
"""

from typing import Callable
from flask_apispec import marshal_with, doc, use_kwargs
from marshmallow import Schema, fields


class ErrorSchema(Schema):
    message = fields.Str(required=True)
    type = fields.Str()
    identifier = fields.Str()


class JobInfoSchema(Schema):
    job_id = fields.Str(required=True)


class JobInputSchema(Schema):
    execute_arguments = fields.Mapping()
    worker_options = fields.Mapping()


class ResourceIdentifierSchema(Schema):
    identifier = fields.Str(required=True)


class ResourceIdentifierListSchema(Schema):
    identifiers = fields.List(fields.Str, required=True)


class EwoksGraphSchema(Schema):
    graph = fields.Mapping()
    nodes = fields.List(fields.Mapping)
    links = fields.List(fields.Mapping)


class EwoksGraphDescriptionSchema(Schema):
    id = fields.Str()
    label = fields.Str()
    category = fields.Str()


class EwoksDataUrlSchema(Schema):
    data_url = fields.Str(required=True)


class EwoksTaskSchema(Schema):
    task_type = fields.Str(required=True)
    task_identifier = fields.Str(required=True)
    category = fields.Str()
    icon = fields.Str()
    required_input_names = fields.List(fields.Str)
    optional_input_names = fields.List(fields.Str)
    output_names = fields.List(fields.Str)


class EwoksGraphListSchema(Schema):
    items = fields.List(fields.Nested(EwoksGraphSchema()))


class EwoksGraphDescriptionListSchema(Schema):
    items = fields.List(fields.Nested(EwoksGraphDescriptionSchema()))


class EwoksTaskListSchema(Schema):
    items = fields.List(fields.Nested(EwoksTaskSchema()))


class DiscoverSchema(Schema):
    modules = fields.List(fields.Str)


class EwoksEventSchema(Schema):
    host_name = fields.Str(required=True)
    process_id = fields.Int(required=True)
    user_name = fields.Str(required=True)
    job_id = fields.Str(required=True)
    binding = fields.Str(dump_default=None, load_default=None)
    context = fields.Str(required=True)
    workflow_id = fields.Str(required=True)
    node_id = fields.Str(dump_default=None, load_default=None)
    task_id = fields.Str(dump_default=None, load_default=None)
    type = fields.Str(required=True)
    time = fields.Str(required=True)
    error = fields.Boolean(dump_default=None, load_default=None)
    error_message = fields.Str(dump_default=None, load_default=None)
    error_traceback = fields.Str(dump_default=None, load_default=None)
    progress = fields.Int(dump_default=None, load_default=None)
    task_uri = fields.Str(dump_default=None, load_default=None)
    input_uris = fields.List(fields.Mapping, dump_default=None, load_default=None)
    output_uris = fields.List(fields.Mapping, dump_default=None, load_default=None)


class EwoksEventQuerySchema(Schema):
    user_name = fields.Str()
    job_id = fields.Str()
    context = fields.Str()
    workflow_id = fields.Str()
    node_id = fields.Str()
    task_id = fields.Str()
    type = fields.Str()
    starttime = fields.Str()
    endtime = fields.Str()
    error = fields.Boolean()


class EwoksEventListSchema(Schema):
    jobs = fields.List(fields.List(fields.Nested(EwoksEventSchema())))


def get_resource_content_schema(resource_type: str):
    if resource_type == "workflow":
        return EwoksGraphSchema
    elif resource_type == "task":
        return EwoksTaskSchema
    elif resource_type == "icon":
        return EwoksDataUrlSchema
    else:
        raise TypeError(resource_type)


def get_resource_content_list_schema(resource_type: str):
    if resource_type == "workflow":
        return EwoksGraphListSchema
    elif resource_type == "task":
        return EwoksTaskListSchema
    else:
        raise TypeError(resource_type)


def get_resource_description_list_schema(resource_type: str):
    if resource_type == "workflow":
        return EwoksGraphDescriptionListSchema
    else:
        raise TypeError(resource_type)


def get_resource(resource_type: str):
    def wrapper(func: Callable):
        bodyschema = get_resource_content_schema(resource_type)
        func = doc(summary=f"Get a {resource_type}")(func)
        func = marshal_with(
            bodyschema,
            code=200,
            description=f"{resource_type} returned",
        )(func)
        func = marshal_with(
            ErrorSchema,
            code=403,
            description=f"no permission to read the {resource_type}",
        )(func)
        func = marshal_with(
            ErrorSchema,
            code=404,
            description=f"requested {resource_type} is not found",
        )(func)
        return func

    return wrapper


def put_resource(resource_type: str):
    def wrapper(func: Callable):
        bodyschema = get_resource_content_schema(resource_type)
        func = doc(summary=f"Update a {resource_type}")(func)
        func = use_kwargs(bodyschema)(func)
        func = marshal_with(
            bodyschema,
            code=200,
            description=f"{resource_type} was overwritten",
        )(func)
        func = marshal_with(
            ErrorSchema,
            code=400,
            description=f"bad {resource_type} update request",
        )(func)
        func = marshal_with(
            ErrorSchema,
            code=403,
            description=f"no permission to write the {resource_type}",
        )(func)
        func = marshal_with(
            ErrorSchema,
            code=404,
            description=f"requested {resource_type} is not found",
        )(func)
        return func

    return wrapper


def post_resource(resource_type: str):
    def wrapper(func: Callable):
        bodyschema = get_resource_content_schema(resource_type)
        func = doc(summary=f"Create a {resource_type}")(func)
        func = use_kwargs(bodyschema)(func)
        func = marshal_with(
            bodyschema,
            code=200,
            description=f"{resource_type} was created",
        )(func)
        func = marshal_with(
            ErrorSchema,
            code=400,
            description=f"bad {resource_type} create request",
        )(func)
        func = marshal_with(
            ErrorSchema,
            code=403,
            description=f"no permission to write the {resource_type}",
        )(func)
        func = marshal_with(
            ErrorSchema,
            code=409,
            description=f"requested {resource_type} already exists",
        )(func)
        return func

    return wrapper


def delete_resource(resource_type: str):
    def wrapper(func: Callable):
        func = doc(summary=f"Delete a {resource_type}")(func)
        func = marshal_with(
            ResourceIdentifierSchema,
            code=200,
            description=f"{resource_type} has been removed or did not exist",
        )(func)
        func = marshal_with(
            ErrorSchema,
            code=403,
            description=f"no permission to delete the {resource_type}",
        )(func)
        return func

    return wrapper


def list_resource_identifiers(resource_type: str):
    def wrapper(func: Callable):
        func = doc(summary=f"Get a list of {resource_type} identifiers")(func)
        func = marshal_with(ResourceIdentifierListSchema, code=200)(func)
        return func

    return wrapper


def list_resource_content(resource_type: str):
    def wrapper(func: Callable):
        func = doc(summary=f"Get a list of {resource_type}s")(func)
        func = marshal_with(get_resource_content_list_schema(resource_type), code=200)(
            func
        )
        return func

    return wrapper


def list_resource_descriptions(resource_type: str):
    def wrapper(func: Callable):
        func = doc(summary=f"Get a list of {resource_type}s")(func)
        func = marshal_with(
            get_resource_description_list_schema(resource_type), code=200
        )(func)
        return func

    return wrapper


def execute_resource(resource_type: str):
    def wrapper(func: Callable):
        func = doc(summary=f"Start {resource_type} execution")(func)
        func = use_kwargs(JobInputSchema)(func)
        func = marshal_with(
            JobInfoSchema,
            code=200,
            description=f"{resource_type} execution started",
        )(func)
        func = marshal_with(
            ErrorSchema,
            code=403,
            description=f"no permission to read the {resource_type}",
        )(func)
        func = marshal_with(
            ErrorSchema,
            code=404,
            description=f"requested {resource_type} is not found",
        )(func)
        return func

    return wrapper


def discover_resources(resource_type: str):
    def wrapper(func: Callable):
        func = doc(summary=f"Discover {resource_type}s")(func)
        func = use_kwargs(DiscoverSchema)(func)
        func = marshal_with(ResourceIdentifierListSchema, code=200)(func)
        func = marshal_with(
            ErrorSchema,
            code=403,
            description=f"no permission to write the {resource_type}",
        )(func)
        func = marshal_with(
            ErrorSchema,
            code=404,
            description="module not found",
        )(func)
        func = marshal_with(
            ErrorSchema,
            code=409,
            description=f"requested {resource_type} already exists",
        )(func)
        return func

    return wrapper


def get_ewoks_events():
    def wrapper(func: Callable):
        func = doc(summary="Get ewoks execution events")(func)
        func = use_kwargs(EwoksEventQuerySchema, location="query")(func)
        func = marshal_with(EwoksEventListSchema, code=200)(func)
        func = marshal_with(
            ErrorSchema,
            code=500,
            description="most likely the server was not configured for ewoks events",
        )(func)
        return func

    return wrapper



from ewokscore import task_discovery
from . import resource
from .. import api


class Task(resource.JsonResource):
    RESOURCE_TYPE = "task"

    def get_identifier(
        self, resource: resource.ResourceContentType
    ) -> resource.ResourceIdentifierType:
        return resource["task_identifier"]

    @api.get_resource("task")
    def get(self, identifier: resource.ResourceIdentifierType) -> resource.ResponseType:
        return self.load_resource(identifier)

    @api.put_resource("task")
    def put(
        self, identifier: resource.ResourceIdentifierType, **resource
    ) -> resource.ResponseType:
        return self.save_resource(
            resource, error_on_missing=True, identifier=identifier
        )

    @api.delete_resource("task")
    def delete(
        self, identifier: resource.ResourceIdentifierType
    ) -> resource.ResponseType:
        return self.delete_resource(identifier)


class Tasks(resource.JsonResource):
    RESOURCE_TYPE = "task"

    def get_identifier(
        self, resource: resource.ResourceContentType
    ) -> resource.ResourceIdentifierType:
        return resource["task_identifier"]

    @api.list_resource_identifiers("task")
    def get(self) -> resource.ResponseType:
        return self.list_resource_identifiers()

    @api.post_resource("task")
    def post(self, **resource) -> resource.ResponseType:
        return self.save_resource(resource, error_on_exists=True)


class Descriptions(resource.JsonResource):
    RESOURCE_TYPE = "task"

    def get_identifier(
        self, resource: resource.ResourceContentType
    ) -> resource.ResourceIdentifierType:
        return resource["task_identifier"]

    @api.list_resource_content("task")
    def get(self):
        return self.list_resource_content()


class DiscoverTasks(resource.JsonResource):
    RESOURCE_TYPE = "task"

    def get_identifier(
        self, resource: resource.ResourceContentType
    ) -> resource.ResourceIdentifierType:
        return resource["task_identifier"]

    @api.discover_resources("task")
    def post(self, modules=None, task_type="class"):
        try:
            tasks = task_discovery.discover_tasks_from_modules(
                *modules, task_type=task_type
            )
        except ModuleNotFoundError as e:
            return self.make_response(404, message=str(e))
        for _resource in tasks:
            self._default_task_properties(_resource)
            response, code = self.save_resource(_resource, error_on_exists=True)
            if code != 200:
                return response, code
        tasks = [desc["task_identifier"] for desc in tasks]
        return self.make_response(200, identifiers=tasks)

    @staticmethod
    def _default_task_properties(_resource: dict) -> None:
        if not _resource.get("icon"):
            _resource["icon"] = "default.png"
        if not _resource.get("label"):
            task_identifier = _resource.get("task_identifier")
            if task_identifier:
                _resource["label"] = task_identifier.split(".")[-1]



def test_single_task(rest_client):
    identifier = "myproject.tasks.Dummy"

    response = rest_client.get(f"/task/{identifier}")
    assert response.status_code == 404

    task1a = {
        "task_identifier": identifier,
        "task_type": "class",
        "required_input_names": ["a"],
    }
    response = rest_client.post("/tasks", json=task1a)
    data = response.get_json()
    assert response.status_code == 200, data
    expected = {
        "required_input_names": ["a"],
        "task_identifier": "myproject.tasks.Dummy",
        "task_type": "class",
    }
    assert data == expected

    response = rest_client.get(f"/task/{identifier}")
    data = response.get_json()
    assert response.status_code == 200, data
    assert data == task1a

    task1b = {
        "task_identifier": identifier,
        "task_type": "class",
        "required_input_names": ["a", "b"],
    }
    response = rest_client.put(f"/task/{identifier}", json=task1b)
    data = response.get_json()
    assert response.status_code == 200, data
    expected = {
        "required_input_names": ["a", "b"],
        "task_identifier": "myproject.tasks.Dummy",
        "task_type": "class",
    }
    assert data == expected

    response = rest_client.get(f"/task/{identifier}")
    data = response.get_json()
    assert response.status_code == 200, data
    assert data == task1b

    response = rest_client.delete(f"/task/{identifier}")
    data = response.get_json()
    assert response.status_code == 200
    assert data == {"identifier": identifier}

    response = rest_client.delete(f"/task/{identifier}")
    data = response.get_json()
    assert response.status_code == 200
    assert data == {"identifier": identifier}

    response = rest_client.get(f"/task/{identifier}")
    data = response.get_json()
    assert response.status_code == 404
    expected = {
        "identifier": identifier,
        "message": f"Task '{identifier}' is not found.",
        "type": "task",
    }
    assert data == expected


def test_multiple_tasks(rest_client, default_task_identifiers):
    response = rest_client.get("/tasks")
    data = response.get_json()
    assert response.status_code == 200
    assert data == {"identifiers": list(default_task_identifiers)}

    task1a = {
        "task_identifier": "myproject.tasks.Dummy1",
        "task_type": "class",
        "required_input_names": ["a"],
    }
    task1b = {
        "task_identifier": "myproject.tasks.Dummy1",
        "task_type": "class",
        "required_input_names": ["a", "b"],
    }
    task2 = {
        "task_identifier": "myproject.tasks.Dummy2",
        "task_type": "class",
        "required_input_names": ["a", "b"],
    }

    response = rest_client.post("/tasks", json=task1a)
    data = response.get_json()
    assert response.status_code == 200, data
    assert data == task1a

    response = rest_client.post("/tasks", json=task1b)
    data = response.get_json()
    assert response.status_code == 409, data
    expected = {
        "identifier": "myproject.tasks.Dummy1",
        "message": "Task 'myproject.tasks.Dummy1' already exists.",
        "type": "task",
    }
    assert data == expected

    response = rest_client.post("/tasks", json=task2)
    data = response.get_json()
    assert response.status_code == 200, data
    assert data == task2

    response = rest_client.get("/tasks")
    data = response.get_json()
    assert response.status_code == 200, data
    expected = set(default_task_identifiers) | {
        "myproject.tasks.Dummy1",
        "myproject.tasks.Dummy2",
    }
    assert set(data["identifiers"]) == expected


def test_discover_tasks(rest_client, default_task_identifiers):
    response = rest_client.get("/tasks")
    data = response.get_json()
    assert response.status_code == 200
    assert data == {"identifiers": list(default_task_identifiers)}

    module = "ewoksserver.tests.dummy_tasks"

    response = rest_client.post("/tasks/discover", json={"modules": [module]})
    data = response.get_json()
    assert response.status_code == 200, data
    expected = {
        "ewoksserver.tests.dummy_tasks.MyTask1",
        "ewoksserver.tests.dummy_tasks.MyTask2",
    }
    assert set(data["identifiers"]) == expected

    response = rest_client.get("/tasks")
    data = response.get_json()
    assert response.status_code == 200
    expected = set(default_task_identifiers) | {
        "ewoksserver.tests.dummy_tasks.MyTask1",
        "ewoksserver.tests.dummy_tasks.MyTask2",
    }
    assert set(data["identifiers"]) == expected

    response = rest_client.post("/tasks/discover", json={"modules": [module]})
    data = response.get_json()
    assert response.status_code == 409, data
    assert "already exists" in data["message"]

    response = rest_client.post("/tasks/discover", json={"modules": ["not_a_module"]})
    data = response.get_json()
    assert response.status_code == 404, data
    assert "No module named" in data["message"]


def test_task_descriptions(rest_client, default_task_identifiers):
    response = rest_client.get("/tasks/descriptions")
    data = response.get_json()
    assert response.status_code == 200
    default_descriptions = [
        desc
        for desc in data["items"]
        if desc["task_identifier"] in default_task_identifiers
    ]
    assert data == {"items": default_descriptions}

    module = "ewoksserver.tests.dummy_tasks"

    response = rest_client.post("/tasks/discover", json={"modules": [module]})
    data1 = response.get_json()
    assert response.status_code == 200, data1

    response = rest_client.get("/tasks/descriptions")
    data2 = response.get_json()["items"]
    data2 = {
        r["task_identifier"] for r in data2 if r["task_identifier"].startswith(module)
    }
    assert response.status_code == 200
    assert set(data1["identifiers"]) == data2



def test_single_workflow(rest_client):
    identifier = "myworkflow"

    response = rest_client.get(f"/workflow/{identifier}")
    assert response.status_code == 404

    workflow1a = {"graph": {"id": identifier}, "nodes": [{"id": "task1"}]}
    response = rest_client.post("/workflows", json=workflow1a)
    data = response.get_json()
    assert response.status_code == 200, data
    assert data == workflow1a

    response = rest_client.get(f"/workflow/{identifier}")
    data = response.get_json()
    assert response.status_code == 200, data
    assert data == workflow1a

    workflow1b = {"graph": {"id": identifier}, "nodes": [{"id": "task2"}]}
    response = rest_client.put(f"/workflow/{identifier}", json=workflow1b)
    data = response.get_json()
    assert response.status_code == 200, data
    assert data == workflow1b

    response = rest_client.get(f"/workflow/{identifier}")
    data = response.get_json()
    assert response.status_code == 200, data
    assert data == workflow1b

    response = rest_client.delete(f"/workflow/{identifier}")
    data = response.get_json()
    assert response.status_code == 200
    assert data == {"identifier": identifier}

    response = rest_client.delete(f"/workflow/{identifier}")
    data = response.get_json()
    assert response.status_code == 200
    assert data == {"identifier": identifier}

    response = rest_client.get(f"/workflow/{identifier}")
    data = response.get_json()
    assert response.status_code == 404
    assert data["message"] == f"Workflow '{identifier}' is not found."


def test_multiple_workflows(rest_client, default_workflow_identifiers):
    response = rest_client.get("/workflows")
    data = response.get_json()
    assert response.status_code == 200
    assert data == {"identifiers": list(default_workflow_identifiers)}

    workflow1a = {"graph": {"id": "myworkflow1"}, "nodes": [{"id": "task1"}]}
    workflow1b = {"graph": {"id": "myworkflow1"}, "nodes": [{"id": "task2"}]}
    workflow2 = {"graph": {"id": "myworkflow2"}, "nodes": [{"id": "task1"}]}

    response = rest_client.post("/workflows", json=workflow1a)
    data = response.get_json()
    assert response.status_code == 200, data

    response = rest_client.post("/workflows", json=workflow1b)
    data = response.get_json()
    assert response.status_code == 409, data
    assert data["message"] == "Workflow 'myworkflow1' already exists."
    response = rest_client.post("/workflows", json=workflow2)
    data = response.get_json()
    assert response.status_code == 200, data

    response = rest_client.get("/workflows")
    data = response.get_json()
    assert response.status_code == 200
    expected = set(default_workflow_identifiers) | {"myworkflow1", "myworkflow2"}
    assert set(data["identifiers"]) == expected


def test_workflow_descriptions(rest_client, default_workflow_identifiers):
    response = rest_client.get("/workflows/descriptions")
    data = response.get_json()
    assert response.status_code == 200
    default_descriptions = [
        desc for desc in data["items"] if desc["id"] in default_workflow_identifiers
    ]
    assert data == {"items": default_descriptions}

    workflow1 = {
        "graph": {"id": "myworkflow1", "label": "label1", "category": "cat1"},
        "nodes": [{"id": "task1"}],
    }
    workflow2 = {"graph": {"id": "myworkflow2"}, "nodes": [{"id": "task1"}]}
    response = rest_client.post("/workflows", json=workflow1)
    data = response.get_json()
    assert response.status_code == 200, data
    response = rest_client.post("/workflows", json=workflow2)
    data = response.get_json()
    assert response.status_code == 200, data

    response = rest_client.get("/workflows/descriptions")
    data = response.get_json()["items"]
    assert response.status_code == 200
    expected = default_descriptions + [
        {"id": "myworkflow1", "label": "label1", "category": "cat1"},
        {"id": "myworkflow2"},
    ]
    data = sorted(data, key=lambda x: x["id"])
    assert data == expected
